module typechecker.

import popeyesig, list, interpreter.

% kind pt 	type.
kind ptpair	type.

type ppriv	Class -> ref -> pt.	% Typing of private variables
				   	% needs to link the variable
					% to the actual object. But 
					% unless constants, we don't
					% need it in programs.

type tclause	pt.			% A clause is a method declaration
type tprog	pt.			% A program is a set of clauses
type tmod	pt.			% An object is contained in a module
type ttest	pt.			% A test is a boolean test.
type theadin	pt.
type theadout 	pt.

type tcmd	pt.			% A cmd is an imperative instruction
type tref	Any -> pt.		% A reference to another object

type none	cmd.			% Empty formula
type dot	pt.			% Yet to be typed
type leaf	cmd.
type tleaf	list ptpair -> pt.

type tp 	Any -> pt -> ptpair.	

type tcheck 	list ptpair -> Any -> list Bny -> pt -> list pt -> o.
type tnodots	list ptpair -> list ptpair -> o.

% Some built-in types
%
type object 	class.
type popbool	pt.
type popint	pt.

% Some typing rules for java constants
%

tcheck Theta null Cont (tref object) TC :-
	tcheck Theta none Cont dot TC.

tcheck Theta (pbool poptrue) Cont popbool TC :-
	tcheck Theta none Cont dot TC.

tcheck Theta (pbool popfalse) Cont popbool TC :-
	tcheck Theta none Cont dot TC.

tcheck Theta (pint I) Cont popint TC :-
	tcheck Theta none Cont dot TC.

% Extension mechanism
%
type extends 	Class -> Dlass -> o.
type inherits	Class -> Dlass -> o.

inherits A B :- extends A B.
inherits A B :- sigma C\ extends A C, inherits C B.

% Type casting
%
tcheck Theta (cast A T) Cont tcmd TC :-
	tcheck ((tp A T)::Theta) none Cont dot TC.

% Axioms
%
tcheck Theta none nil dot nil.
tcheck Theta leaf nil (tleaf Theta) nil. 

% Transfer
%
tcheck Theta none (B::Cont) dot (TB::TC) :-
	tcheck Theta B Cont TB TC.

% Type-check
%
tcheck Theta A Cont TA TC :-
	lmemb Theta (tp A T) top, T = TA,
	print "t-check\n",
	tcheck Theta none Cont dot TC.

% Extension
%
tcheck Theta A Cont (tref TA) TC :-
	lmemb Theta (tp A (tref TB)) top, inherits TB TA,
	print "extension\n",
	tcheck Theta none Cont dot TC.

% Init
%
tcheck Theta ( init A N ) Cont tcmd TC :-
	tcheck Theta A nil Tau nil,
	tcheck Theta N nil Tau nil,
	print "init\n",
	tcheck Theta none Cont dot TC.

% Rw-write	
%
tcheck Theta ( rw A O N ) Cont tcmd TC :-	
	tcheck Theta A nil Tau nil,
	tcheck Theta O nil dot nil,
	tcheck Theta N nil Tau nil,
	print "rw-write\n",
	tcheck ((tp O Tau)::Theta) none Cont dot TC.	

% Rw-read	
%
tcheck Theta ( rw A O N ) Cont tcmd TC :-	
	tcheck Theta A nil Tau nil,
	tcheck Theta O nil dot nil,
	tcheck Theta N nil dot nil,
	print "rw-read\n",
	tcheck ((tp O Tau)::(tp N Tau)::Theta) none Cont dot TC.

% Rw-test	
%
tcheck Theta ( rw A O N ) Cont ttest TC :-	
	tcheck Theta A nil Tau nil,
	tcheck Theta O nil Tau nil,
	tcheck Theta N nil Tau nil,
	print "rw-test\n",
	tcheck Theta none Cont dot TC.

% ==
%
tcheck Theta (A == B) Cont tcmd TC :-
	tcheck Theta A nil dot nil,
	tcheck Theta B nil Tau nil,
	print "==\n",
	tcheck ((tp A Tau)::Theta) none Cont dot TC.

tcheck Theta (A == B) Cont ttest TC :-
	print "1",
	tcheck Theta A nil Tau nil,
	tcheck Theta B nil Tau nil,
	print "==\n",
	tcheck Theta none Cont dot TC.

tcheck Theta (A == B) Cont ttest TC :-
	print "2",
	tcheck Theta B nil Tau nil,
	tcheck Theta A nil Tau nil,
	print "==\n",
	tcheck Theta none Cont dot TC.


% Empty
%
tcheck Theta empty Cont tcmd TC :-
	print "empty\n",
	tcheck Theta none Cont dot TC.

tcheck Theta empty Cont ttest TC :-
	print "DUMMY test\n",
	tcheck Theta none Cont dot TC.
	

% Popnew
%
tcheck Theta (popnew Class S P) Cont tcmd TC :-
	tcheck Theta (new Class P) Cont tcmd TC.

% Branch
%
tcheck Theta (A +> B |> C) Cont tcmd TC :-
	tcheck Theta A nil ttest nil,
	tcheck Theta B (leaf::nil) tcmd ((tleaf LeafB)::nil),
	tcheck Theta C (leaf::nil) tcmd ((tleaf LeafC)::nil), 
		tnodots LeafB NDLeafB,
		tnodots LeafC NDLeafC,
			append NDLeafB NDLeafC LeafBC top,
			append LeafBC Theta ThetaLeaf top,
	print "branch\n",
 	tcheck ThetaLeaf none Cont dot TC. 

tnodots nil nil.
tnodots ((tp M dot)::Theta) NewTheta & 
tnodots ((tp M T)::Theta) ((tp M T)::NewTheta) :- !, tnodots Theta NewTheta. 

% Cont
%
tcheck Theta (A >> B) Cont tcmd TC :-
	print ">>\n",
	tcheck Theta A (B::Cont) tcmd (tcmd::TC).


% Clause
%
tcheck Theta (A >- B) Cont tclause TC :-
	print "clause\n",
	tcheck Theta A (B::A::Cont) theadin (tcmd::theadout::TC).

tcheck Theta (poppi S A) Cont tclause TC :-
	print "pi ",
	pi C\ tcheck ((tp C dot)::Theta) (A C) Cont tclause TC. 

% Prog
%
tcheck Theta (A :: B) Cont tprog TC :-
	print "prog\n",		
	tcheck Theta A nil tclause nil,
	tcheck Theta B Cont tprog TC.

tcheck Theta nil Cont tprog TC :-
	tcheck Theta none Cont dot TC.

% Private
%
tcheck Theta (loc C D P) Cont tprog TC :-
	print "loc ",
	tcheck Theta (ppriv C D) nil Tau nil,
	pi E\ tcheck ((tp E Tau)::Theta) (P E) Cont tprog TC.

tcheck Theta (poploc S C D P) Cont tprog TC :-
	tcheck Theta (loc C D P) Cont tprog TC.

% Constant
%
tcheck Theta (sta C N V P) Cont tprog TC :-
	print "sta ",
	tcheck Theta V nil Tau nil,
	tcheck ((tp (pcst C N) Tau)::Theta) P Cont tprog TC. 

tcheck Theta (popsta S C N V P) Cont tprog TC :-
	tcheck Theta (sta C N V P) Cont tprog TC.

% Context
%
tcheck Theta (popcon Lin Prog) Cont tprog TC :-
	print "context\n",
	tcheck Theta Prog Cont tprog TC.









